#!/usr/bin/python

## expendable

## Copyright (C) 2008 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gtk
import locale
import math
import pango

class Graph (gtk.DrawingArea):
    def __init__ (self, datasets=[], error=None):
        gtk.DrawingArea.__init__ (self)
        self.set_datasets (datasets, error)
        self.connect ('configure-event', self.configure_event)
        self.connect ('expose-event', self.expose_event)
        self.connect ('button-press-event', self.button_press_event)
        self.connect ('motion-notify-event', self.button_press_event)
        self.add_events (gtk.gdk.BUTTON_PRESS_MASK |
                         gtk.gdk.BUTTON_MOTION_MASK)
        self.set_size_request (200, 200)

    def set_datasets (self, datasets, error=None):
        self.datasets = datasets
        self.error = error
        colors = {}
        for dataset in datasets:
            try:
                spec = dataset['color']
                c = gtk.gdk.color_parse (spec)
                colors[spec] = c
            except KeyError:
                pass
            except ValueError:
                print "Unable to locate color %s" % spec

        self.colors = colors
        self.special_point = None
        self.queue_resize ()

    def button_press_event (self, widget, event):
        (left, width, max_x, bottom, height, max_y) = self.widget_boundaries
        x_click = round (max_x * (event.x - left) / width)
        y_click = round (max_y * (bottom - event.y) / height)
        self.special_point = None
        if x_click >= 0 and x_click <= max_x:
            best = None
            for dataset in self.datasets:
                name = dataset['name']
                for x, y in dataset['data']:
                    if x == x_click:
                        if best == None:
                            best = (name, y, abs (y - y_click))
                        else:
                            y_distance = abs (y - y_click)
                            if y_distance < best[2]:
                                best = (name, y, y_distance)

            if best != None:
                self.special_point = (x_click, best[1], best[0])

        self.queue_resize ()
        return False

    def configure_event (self, widget, event):
        x, y, w, h = widget.get_allocation ()
        self.pixmap = gtk.gdk.Pixmap (widget.window, w, h)
        self.gc = {}
        for name, color in self.colors.iteritems ():
            c = self.pixmap.new_gc ()
            c.set_rgb_fg_color (color)
            self.gc[name] = c
        self.layout = pango.Layout (widget.create_pango_context ())
        self.canvas_width = w
        self.canvas_height = h
        self.plot ()
        return True

    def expose_event (self, widget, event):
        x, y, w, h = event.area
        widget.window.draw_drawable (widget.get_style ().fg_gc[gtk.STATE_NORMAL],
                                     self.pixmap, x, y, x, y, w, h)
        return False

    def plot (self):
        # Clear the pixmap.
        self.pixmap.draw_rectangle (self.get_style ().white_gc,
                                    True, 0, 0,
                                    self.canvas_width, self.canvas_height)

        if len (self.datasets) == 0:
            # Nothing to plot.
            return

        ## Legend
        # Find the longest legend label width.
        max_width = 1
        for dataset in self.datasets:
            name = dataset['name']
            self.layout.set_text (name)
            extents = self.layout.get_pixel_extents ()[1]
            width = extents[2]
            if width > max_width:
                max_width = width

        self.layout.set_text ('l')
        line_height = self.layout.get_pixel_extents ()[1][3] + 3
        legendleft = self.canvas_width - (12 + 6 + max_width + 12)
        legendtop = 12
        for dataset in self.datasets:
            name = dataset['name']
            try:
                c = dataset['color']
                gc = self.gc[c]
            except KeyError:
                gc = self.get_style ().black_gc

            self.layout.set_text (name)
            extents = self.layout.get_pixel_extents ()[1]
            midy = legendtop + (extents[3] / 2)
            self.pixmap.draw_lines (gc,
                                    [(legendleft, midy),
                                     (legendleft + 12, midy)])

            self.pixmap.draw_layout (self.get_style ().black_gc,
                                     legendleft + 12 + 6, legendtop,
                                     self.layout)

            legendtop += line_height

        legendbottom = legendtop

        # Work out how our axes need to be, and what space they need
        # for labels.
        max_x = 1
        max_y = 1
        for dataset in self.datasets:
            data = dataset['data']
            for x, y in data:
                if x > max_x:
                    max_x = x
                if y > max_y:
                    max_y = y

        def axis_max (f, base):
            a = pow (base, int (math.log (f) / math.log (base))) / 2
            if a == 0:
                a = 1
            return a, a * (1 + int (f/a))

        orig_max_x = max_x
        x_step, max_x = axis_max (orig_max_x, 12)
        if x_step >= 12:
            x_step, max_x = axis_max (orig_max_x / 12, 10)
            x_step *= 12
            max_x *= 12
        y_step, max_y = axis_max (max_y, 10)

        self.layout.set_text ("%d" % max_y)
        extents = self.layout.get_pixel_extents ()[1]
        margin = 12
        leftmargin = extents[2] + 6 + margin
        topmargin = margin
        bottommargin = extents[3] + 6 + 12
        rightmargin = margin * 2

        ## Draw the graph.
        left = leftmargin
        top = topmargin
        right = self.canvas_width - rightmargin
        bottom = self.canvas_height - bottommargin
        width = right - left
        height = bottom - top

        # Check to see whether the legend can be overlaid on the plot.
        for dataset in self.datasets:
            data = dataset['data']
            for x, y in data:
                plotx = left + int (x * width / max_x)
                ploty = bottom - int (y * height / max_y)
                if (plotx > legendleft and
                    ploty < legendbottom):
                    # No space.
                    right = legendleft - rightmargin
                    width = right - left

        if left >= right or bottom <= top:
            # No space to draw the plot!
            return

        # Draw axes.
        self.pixmap.draw_lines (self.get_style ().black_gc,
                                [(left, bottom + 5),
                                 (left, top)])
        self.pixmap.draw_lines (self.get_style ().black_gc,
                                [(left - 5, bottom),
                                 (right, bottom)])
        # If we need to draw attention to a particular x coordinate,
        # draw in a vertical line.
        if self.error != None:
            color = 'red'
            try:
                c = self.gc[color]
            except KeyError:
                c = self.pixmap.new_gc ()
                c.set_rgb_fg_color (color)
                self.gc[color] = c

            x = left + int (self.error * width / max_x)
            self.pixmap.draw_lines (c,
                                    [(x, bottom + 15),
                                     (x, top)])

        # Draw axis marks.
        if x_step < 12:
            x_step = 1

        last_label_left = None
        for i in xrange (max_x, 0, -x_step):
            x = left + int (i * width / max_x)

            ysize = 2
            xsize = 1
            years = i / 12
            months = i % 12
            if i != 0 and months == 0:
                ysize *= 2
                xsize += 1

            if i != 0 and months == 0:
                label = "%dy" % years
                self.layout.set_text (label)
                extents = self.layout.get_pixel_extents ()[1]

                label_right = x + (extents[2] / 2)
                if last_label_left == None or label_right < last_label_left:
                    last_label_left = x - (extents[2] / 2)
                    self.pixmap.draw_layout (self.get_style ().black_gc,
                                             last_label_left,
                                             bottom + 6,
                                             self.layout)
                    ysize *= 2

            self.pixmap.draw_rectangle (self.get_style ().black_gc, True,
                                        x, bottom, xsize, ysize)

        last_label_bottom = None
        for i in xrange (max_y, 0, -y_step):
            y = bottom - int (i * height / max_y)

            size = 2
            label = "%d" % i
            self.layout.set_text (label)
            extents = self.layout.get_pixel_extents ()[1]
            label_left = left - extents[2] - 6
            label_top = y - (extents[3] / 2)
            if last_label_bottom == None or label_top > last_label_bottom:
                self.pixmap.draw_layout (self.get_style ().black_gc,
                                         label_left,
                                         label_top,
                                         self.layout)
                last_label_bottom = y + (extents[3] / 2)
                size *= 2

            self.pixmap.draw_lines (self.get_style ().black_gc,
                                    [(left - size, y),
                                     (left, y)])

        for dataset in self.datasets:
            points = []
            data = dataset['data']
            try:
                c = dataset['color']
                gc = self.gc[c]
            except KeyError:
                gc = self.get_style ().black_gc

            # Draw vertical lines at the beginning and end if
            # necessary.
            n = len (data)
            if n > 0:
                x, y = data[0]
                if x > 0:
                    data.insert (0, (x, 0))

                x, y = data[len (data) - 1]
                if x < orig_max_x and y > 0:
                    data.append ((x, 0))

            for x, y in data:
                points.append ((left + int (x * width / max_x),
                                bottom - int (y * height / max_y)))
            if len (points) > 0:
                self.pixmap.draw_lines (gc, points)

        self.widget_boundaries = (left, width, max_x, bottom, height, max_y)
        if self.special_point != None:
            (x, y, name) = self.special_point
            self.layout.set_text (locale.currency (y, True, True))
            x = left + int (x * width / max_x)
            y = bottom - int (y * height / max_y)
            w = 4
            self.pixmap.draw_arc (self.get_style ().black_gc, True,
                                  x - w / 2, y - w / 2, w, w, 0, 64 * 360)
            extents = self.layout.get_pixel_extents ()[1]
            w = extents[2]
            self.pixmap.draw_rectangle (self.get_style ().white_gc, True,
                                        x - w / 2, y + 3,
                                        w, extents[3])
            self.pixmap.draw_layout (self.get_style ().black_gc,
                                     x - w / 2, y + 3, self.layout)

if __name__ == '__main__':
    g = Graph ([{'name': 'Test',
                 'data': [(0, 2000), (1, 1900), (2, 1500), (3, 500), (4, 0)]}])
    w = gtk.Window ()
    w.add (g)
    w.show_all ()
    w.connect ("delete_event", gtk.main_quit)
    gtk.main ()
